package gjocodonlib;
#
##############################################################################
#
#  Copyright (c) 2009 Gary J. Olsen
#
#  This is free software. You can redistributeit and/or modify it under
#  the terms of the GNU General Public License version 3.
#
#  Copies of the license are available from the Free Software Foundation:
#
#     http://www.gnu.org/licenses
#
##############################################################################
#
#   \%counts        = entry_codon_count( [ \%counts, ] @seq_entry_ref )
#   \%counts        = seq_codon_count( [ \%counts, ] $sequence )
#
#    @counts        = codon_count_list( \%counts [,  @order ] )
#    @counts        = codon_count_list( \%counts [, \@order ] )
#
#   \@counts        = codon_count_package( \%counts )
#   \@counts        = codon_count_package_20( \%counts )
#
#   \@counts        = split_counts( $codon_count_string )
# ( \@counts, $id ) = split_counts( $codon_count_string )
#
#   \@total_counts  = sum_counts( \@per_gene_count_arrays )
#   \@total_counts  = sum_counts( \@gene_1_counts, \@gene_2_counts, ... )
#   \%total_counts  = sum_counts( \@per_gene_count_hashes )
#   \%total_counts  = sum_counts( \%gene_1_counts, \%gene_2_counts, ... )
#
#                     report_counts(       \@packaged_counts )
#                     report_counts(       \@packaged_counts, $id )
#                     report_counts( \*FH, \@packaged_counts )
#                     report_counts( \*FH, \@packaged_counts, $id )
#
#   \@freqs                = count_to_freq( \@counts [, $pseudocount ] )
#   \@freqs                = count_to_freq( \%counts [, $pseudocount ] )
#
#   \@freqs                = split_frequencies( $codon_freq_string )
# ( \@freqs, $scr, $desc ) = split_frequencies( $codon_freq_string )
#
#   $codon_freq_string = frequencies_as_string( \@freqs )
#
#                        report_frequencies(       \@freqs )
#                        report_frequencies( \*FH, \@freqs )
#
#   @chisqr_df_n       = codon_usage_chi_sqr( \%freqs, \%cnt1 [ , ... ] )
#
# ( $chisqr, $df, $n ) = count_vs_count_chi_sqr( \@cnt_1, \@cnt_2 )
# ( $chisqr, $df, $n ) = count_vs_count_chi_sqr( \%cnt_1, \%cnt_2 )
#
# ( $chisqr, $df, $n ) = count_vs_freq_chi_sqr( \@counts, \@freqs )
# ( $chisqr, $df, $n ) = count_vs_freq_chi_sqr( \%counts, \@freqs )
#
#   @p_values          = codon_usage_p_values( \@counts, \@freq_sets, \%opts )
#
# Functions that evaluates the total score for counts of multiple genes.
# Do it entirely in perl:
#
#   $score = codon_freq_score( \@freq, \@per_gene_counts, \%options )
#   $score = codon_freq_score( \@freq, \%per_gene_counts, \%options )
#
# Evaluate mulitple sets of frequencies using an external evalution process.
# Counts can be in file, arg list, or options.
#
#   @scored_freqs = score_codon_frequencies( \@freq_sets, \@counts, \%options )
#  \@scored_freqs = score_codon_frequencies( \@freq_sets, \@counts, \%options )
#   @scored_freqs = score_codon_frequencies( \@freq_sets,           \%options )
#  \@scored_freqs = score_codon_frequencies( \@freq_sets,           \%options )
#
#   $score = codon_freq_score( \@freq, \@per_gene_counts, \%options )
#   $score = codon_freq_score( \@freq, \%per_gene_counts, \%options )
#
#
#   @per_gene_aa_cnt      = codon_counts_2_aa_counts( @per_gene_codon_cnt )
#
#   @per_gene_codon_cnt   = simulate_genome( \@packaged_freqs, @per_gene_aa_cnt )
#
#            \@modal_freqs   = modal_codon_usage( \@gene_cnt_pkgs, \%options )
#  ( $score, \@modal_freqs ) = modal_codon_usage( \@gene_cnt_pkgs, \%options )
#
#   $distance = codon_freq_distance( \@freq1, \@freq2, $type ) # D = type 2
#   $distance = codon_freq_distance_1( \@freq1, \@freq2 ) # Euclidian over all codons
#   $distance = codon_freq_distance_2( \@freq1, \@freq2 ) # Manhatten within aa, and Euclidian over aas (recommended)
#   $distance = codon_freq_distance_3( \@freq1, \@freq2 ) # Manhatten over all codons
#
#   @projections = project_on_freq_vector_by_dist( \@freq_0, \@freq_1,   \@freq1, \@freq2, ...   )
#   @projections = project_on_freq_vector_by_dist( \@freq_0, \@freq_1, [ \@freq1, \@freq2, ... ] )
#               # each projection is [ $position, $distance_from_axis ]
#
#   @projections = project_on_freq_vector_by_chi_sqr( \@freq_0, \@freq_1,   \@cnts1, \@cnts2, ...   )
#   @projections = project_on_freq_vector_by_chi_sqr( \@freq_0, \@freq_1, [ \@cnts1, \@cnts2, ... ] )
#               # each projection is [ $position, $chi_sqr, $df, $ncodon ]
#
#-------------------------------------------------------------------------------
#  Some earlier versions:
#
#  \@freqs             = packaged_count_to_freq( \@counts [, $pseudocount ] )
#  \@total_counts      = sum_packaged_counts( \@per_gene_pakaged_counts )
# ( $chisqr, $df, $n ) = codon_usage_pairwise_chi_sqr( \%cnt1, \%cnt2 )
# ( $chisqr, $df, $n ) = count_package_chi_sqr( \@cnt_pkg1, \@cnt_pkg2 )
# ( $chisqr, $df, $n ) = packaged_codon_usage_chi_sqr( \@freqs, \@cnt )
#
#===============================================================================

use strict;
use Carp qw( croak );
# use Data::Dumper qw( Dumper );

use IPC::Open2 qw( open2 );

use gjoseqlib qw(
        to_DNA_seq
        pack_seq
        @aa_n_codon_order
        %genetic_code
        %n_codon_for_aa
        %amino_acid_codons_DNA
        );

use gjostat qw(
        chi_square
        contingency_chi_sqr_2
        );

#  Functions used for simulating codon usage:

use gjosegmentlib qw(
       segment_new_tree
       segment_by_coord
       );

use Data::Dumper;

our @aa_order;
our @codon_order;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw(
        entry_codon_count
        seq_codon_count
        codon_count_list
        count_vs_count_chi_sqr
        count_vs_freq_chi_sqr
        codon_usage_chi_sqr
        codon_count_package
        codon_count_package_20
        sum_counts
        count_to_freq
        codon_usage_p_values
        codon_freq_score
        report_counts
        split_counts
        report_frequencies
        frequencies_as_string
        split_frequencies
        codon_counts_2_aa_counts
        simulate_genome

        modal_codon_usage
        score_codon_frequencies

        codon_freq_distance
        codon_freq_distance_1
        codon_freq_distance_2
        codon_freq_distance_3

		project_on_freq_vector_by_dist
		project_on_freq_vector_by_chi_sqr

        packaged_count_to_freq
        sum_packaged_counts
        codon_usage_pairwise_chi_sqr
        count_package_chi_sqr
        packaged_codon_usage_chi_sqr
        );

our @EXPORT_OK = qw(
        @aa_order
        @codon_order
        );



#-----------------------------------------------------------------------------
#  Compile codon usage of one or more sequence entries.  Skip initiator, and
#  the last codon, if it is a terminator.  If the routine is passed a hash,
#  add to it.  If it is not, create a new hash.
#
#     \%counts = entry_codon_count( [ \%counts, ] @seq_entry_ref )
#
#-----------------------------------------------------------------------------
sub entry_codon_count
{
    my $cnt = ( ref( $_[0] ) eq 'HASH' ) ? shift : {};
    my $seq;

    foreach ( @_ ) {
        ( ref( $_ ) eq 'ARRAY' ) && defined( $seq = $_->[2] ) || next;
        seq_codon_count( $cnt, $seq );
    }

    return $cnt;
}


#-----------------------------------------------------------------------------
#  Compile codon usage for a sequence.  Skip initiator, and the last codon
#  if it is a terminator.  If the routine is passed a hash, add to it.
#  Otherwise, create a new hash.
#
#     \%counts = seq_codon_count( [ \%counts, ] $sequence )
#
#-----------------------------------------------------------------------------
sub seq_codon_count
{
    my $cnt = ( ref( $_[0] ) eq 'HASH' ) ? shift : {};
    my $seq = shift;
    defined( $seq ) || return undef;

    #  Pack, convert U to T, skip first 3 residues, convert to upper case
    #  and extract valid triplets:

    my @codons = grep { /^[ACGT][ACGT][ACGT]$/ }
                 substr( uc to_DNA_seq( pack_seq( $seq ) ), 3 ) =~ m/.../g;

    #  Dump last codon if terminator:

    if ( $genetic_code{ $codons[ -1 ] } eq "*" ) { pop @codons }

    #  Compile the codons:

    foreach ( @codons ) { $cnt->{ $_ }++ }

    return $cnt;
}


#  Make lists in order of codons per amino acid:
#
#  @aa_n_codon_order  = qw( L R S A G P T V I C D E F H K N Q Y M W );

@aa_order    = @aa_n_codon_order;
@codon_order = map { @{ $amino_acid_codons_DNA{ $_ } } } @aa_order;

#-----------------------------------------------------------------------------
#  Convert a codon count in a hash to a list in standard order.
#
#     @counts = codon_count_list( \%counts [,  @order ] )
#     @counts = codon_count_list( \%counts [, \@order ] )
#
#-----------------------------------------------------------------------------
sub codon_count_list
{
    ( ref( $_[0] ) eq 'HASH' ) || return undef;
    my $hash = shift;
    if ( ( @_ == 1 ) && ref( $_[0] ) eq 'ARRAY' ) { @_ = @{ $_[0] } }
    if   ( @_ <  2 )                              { @_ = @codon_order }

    map { defined( $hash->{ $_ } ) ? $hash->{ $_ } : 0 } @_;
}


#-----------------------------------------------------------------------------
#  Compare two codon usages by chi-square.
#
#     ( $chisqr, $df, $n ) = codon_usage_pairwise_chi_sqr( \%cnt1, \%cnt2 )
#
#-----------------------------------------------------------------------------
sub codon_usage_pairwise_chi_sqr
{
    my ( $cnt1, $cnt2 ) = @_;
    ref( $cnt1 ) eq 'HASH' && ref( $cnt2 ) eq 'HASH'
        || die "codon_usage_2_chi_sqr requires two HASH references\n";

    my ( $chisqr, $df, $total ) = (0, 0, 0);

    foreach my $aa ( qw( A C D E F G H I K L N P Q R S T V Y ) )
    {
        my @codons = @{ $amino_acid_codons_DNA{ $aa } };
        my ($c, $d, $n) = contingency_chi_sqr_2( [ map { $cnt1->{ $_ } } @codons ],
                                                 [ map { $cnt2->{ $_ } } @codons ]
                                               );
        if ( $d > 0 ) { $chisqr += $c; $df += $d; $total += $n }
    }

    ( $chisqr, $df, $total )
}


#-----------------------------------------------------------------------------
#  Compare codon usage(s) to expected frequencies by chi-square.
#
#     @[ $chisqr, $df, $n ] = codon_usage_chi_sqr( \%freqs, \%cnt1 [ , ... ] )
#
#-----------------------------------------------------------------------------
sub codon_usage_chi_sqr
{
    ( @_ > 1 ) || die "Usage: codon_usage_chi_sqr( \%freqs, \%cnts ... )\n";
    my $freqs = shift;
    ref( $freqs ) eq 'HASH'
        || die "codon_usage_chi_sqr args must be HASH references\n";

    my @out = ();

    foreach my $cnts ( @_ ) {
        ref( $cnts ) eq 'HASH'
            || die "codon_usage_chi_sqr args must be HASH references\n";
        my ( $chisqr, $df, $total ) = (0, 0, 0);

        foreach my $aa ( qw( A C D E F G H I K L N P Q R S T V Y ) )
        {
            my @codons = @{ $amino_acid_codons_DNA{ $aa } };
            my ($c, $d, $n) = gjostat::chi_square( [ map { $freqs->{ $_ } } @codons ],
                                                   [ map { $cnts->{ $_ }  } @codons ]
                                                 );
            if ( $d > 0 ) { $chisqr += $c; $df += $d; $total += $n }
        }

        push @out, [ $chisqr, $df, $total ];
    }

    @out
}


#-----------------------------------------------------------------------------
#  Convert a codon count hash to a list of counts bundled by amino acid.
#  Default version omits M and W.
#
#     \@counts = codon_count_package( \%counts )
#
#  \@counts = [ [ n(GCA), n(GCG), n(GCT), n(GCC) ], # A
#               [ n(TGT), n(TGC) ],                 # C
#               [ n(GAT), n(GAC) ],                 # D
#               .
#               .
#               .
#               [ n(TAT), n(TAC) ]                  # Y
#             ];
#-----------------------------------------------------------------------------
sub codon_count_package
{
    my $cnts = shift;
    ref( $cnts ) eq 'HASH' or return undef;

    [ map { [ map { $cnts->{ $_ } || 0                  # map codon to counts
                  } @{ $amino_acid_codons_DNA{ $_ } }   # for each codon
            ]
          } qw( A C D E F G H I K L N P Q R S T V Y )   # for 18 amino acids
    ]
}


#-----------------------------------------------------------------------------
#  Convert a codon count hash to a list of counts bundled by amino acid.
#
#     \@counts = codon_count_package_20( \%counts )
#
#  \@counts = [ [ n(GCA), n(GCG), n(GCT), n(GCC) ], # A
#               [ n(TGT), n(TGC) ],                 # C
#               [ n(GAT), n(GAC) ],                 # D
#               .
#               .
#               .
#               [ n(TAT), n(TAC) ]                  # Y
#               [ n(ATG) ]                          # M
#               [ n(TGA) ]                          # W
#             ];
#-----------------------------------------------------------------------------
sub codon_count_package_20
{
    my $cnts = shift;
    ref( $cnts ) eq 'HASH' or return undef;

    [ map { [ map { $cnts->{ $_ } || 0                  # map codon to counts
                  } @{ $amino_acid_codons_DNA{ $_ } }   # for each codon
            ]
          } qw( A C D E F G H I K L N P Q R S T V Y M W ) # for 20 amino acids
    ]
}


#-----------------------------------------------------------------------------
#  Sum codon counts across genes.  Handles any nested arrays of counts.
#
#     \@total_counts = sum_packaged_counts( \@per_gene_pakaged_counts )
#
#-----------------------------------------------------------------------------
sub sum_packaged_counts
{
    my $gene_counts = shift;
    ref( $gene_counts ) eq 'ARRAY'                     # array of genes
        and ref( $gene_counts->[0] ) eq 'ARRAY'        # array of amino acids
        and ref( $gene_counts->[0]->[0] ) eq 'ARRAY'   # array of codons
        or return undef;

    my @ttl_cnt;
    my $gene_data;
    foreach $gene_data ( @$gene_counts )
    {
        my $i = 0;
        my $gene_aa_data;
        foreach $gene_aa_data ( @$gene_data )
        {
             my $ttl_aa_cnts = $ttl_cnt[ $i++ ] ||= [];
             my $j = 0;
             my $gene_codon_cnt;
             foreach $gene_codon_cnt ( @$gene_aa_data )
             {
                 $ttl_aa_cnts->[ $j++ ] += $gene_codon_cnt;
             }
        }
    }

    \@ttl_cnt;
}


#-----------------------------------------------------------------------------
#  Sum codon counts across genes.
#
#     \@total_counts = sum_counts( \@per_gene_count_arrays )
#     \@total_counts = sum_counts( \@gene_1_counts, \@gene_2_counts, ... )
#
#     \%total_counts = sum_counts( \@per_gene_count_hashes )
#     \%total_counts = sum_counts( \%gene_1_counts, \%gene_2_counts, ... )
#
#-----------------------------------------------------------------------------
sub sum_counts
{
    return undef if ( ! @_ ) || ( ( @_ == 1 ) && ( ref $_[0] ne 'ARRAY' ) );

    #  Are counts in arrays?

    if ( ref $_[0] eq 'ARRAY' && ref $_[0]->[0] eq 'ARRAY' )
    {
        my @ttl_cnt;
        foreach my $gene_data ( ref $_[0]->[0]->[0] eq 'ARRAY' ? @{$_[0]} : @_ )
        {
            my $i = 0;
            foreach my $gene_aa_data ( @$gene_data )
            {
                my $ttl_aa_cnts = $ttl_cnt[ $i++ ] ||= [];
                my $j = 0;
                foreach my $gene_codon_cnt ( @$gene_aa_data )
                {
                    $ttl_aa_cnts->[ $j++ ] += $gene_codon_cnt;
                }
            }
        }
        return \@ttl_cnt;
    }

    #  Are counts in hash?

    if ( ref $_[0] eq 'HASH' || ref $_[0]->[0] eq 'HASH' )
    {
        my %ttl_cnt;
        foreach my $gene_data ( ref $_[0] eq 'HASH' ? $_ : @{$_[0]} )
        {
            foreach ( keys %$gene_data ) { $ttl_cnt{ $_ } += $gene_data->{ $_ } }
        }
        return \%ttl_cnt;
    }

    #  I don't understand args

    return undef;
}


#-----------------------------------------------------------------------------
#  Compare two codon usages by chi-square.
#
#     ( $chisqr, $df, $n ) = count_package_chi_sqr( \@cnt_pkg1, \@cnt_pkg2 )
#
#-----------------------------------------------------------------------------
sub count_package_chi_sqr
{
    my ( $cnt1, $cnt2 ) = @_;
    ref( $cnt1 ) eq 'ARRAY' && ref( $cnt2 ) eq 'ARRAY'
        || die "count_package_chi_sqr requires two ARRAY references\n";
    ( @$cnt1 == @$cnt2 ) or die "count_package_chi_sqr requires arrays of equal size\n";

    my @c2 = @$cnt2[ 0 .. 17 ];
    my ( $chisqr, $df, $total ) = (0, 0, 0);

    foreach my $c1 ( @$cnt1[0 .. 17] )
    {
        my ($c, $d, $n) = contingency_chi_sqr_2( $c1, shift @c2 );
        if ( $d && $n ) { $chisqr += $c; $df += $d; $total += $n }
    }

    ( $chisqr, $df, $total )
}


#-----------------------------------------------------------------------------
#  Compare two codon counts by chi-square.
#
#     ( $chisqr, $df, $n ) = count_vs_count_chi_sqr( \@cnt_1, \@cnt_2 )
#     ( $chisqr, $df, $n ) = count_vs_count_chi_sqr( \%cnt_1, \%cnt_2 )
#
#-----------------------------------------------------------------------------
sub count_vs_count_chi_sqr
{
    my ( $cnt1, $cnt2 ) = @_;

    $cnt1 = codon_count_package( $cnt1 ) if ref $cnt1 eq 'HASH';
    $cnt2 = codon_count_package( $cnt2 ) if ref $cnt2 eq 'HASH';

    return () if ref $cnt1 ne 'ARRAY' || ref $cnt2 ne 'ARRAY' || ( @$cnt1 != @$cnt2 );

    my ( $chisqr, $df, $total ) = (0, 0, 0);
    for ( my $i = 0; $i <= 17; $i++ )
    {
        my ( $c, $d, $n ) = contingency_chi_sqr_2( $cnt1->[$i], $cnt2->[$i] );
        if ( $d && $n ) { $chisqr += $c; $df += $d; $total += $n }
    }

    ( $chisqr, $df, $total )
}


#-----------------------------------------------------------------------------
#  Convert packaged counts to packaged frequencies.
#  Optionally add a pseudocount (e.g., 1) to each group.
#  The number of amino acids is trimmed to 18.
#
#     \@freqs = packaged_count_to_freq( \@counts [, $pseudocount ] )
#
#     \@counts = [ [ n1, n2, n3, n4 ], [ n5, n6 ], ... [ nn1, nn2 ] ];
#     \@freqs  = [ [ f1, f2, f3, f4 ], [ nf, f6 ], ... [ fn1, fn2 ] ];
#-----------------------------------------------------------------------------
sub packaged_count_to_freq
{
    my $cnts = shift;
    ref( $cnts ) eq 'ARRAY' or return undef;
    my $pseudocnt = ( shift ) || 0;

    [ map { my $i = 0;                          # elements in group
            my $n = $pseudocnt;
            foreach ( @$_ ) { $n += $_ ; $i++ } # total count in group
            $n ||= 1;                           #   or 1
            my $pci = $i ? $pseudocnt / $i : 0; # and per element pseudocount
            [ map { ( $_ + $pci ) / $n } @$_ ]  # used to make frequencies
          } @$cnts[ 0 .. 17 ]                   # for each group of counts
    ]
}


#-----------------------------------------------------------------------------
#  Convert counts to packaged frequencies.  
#  Optionally add a pseudocount (e.g., 1) to each group.
#  The number of amino acids is trimmed to 18.
#
#     \@freqs = count_to_freq( \@counts [, $pseudocount ] )
#     \@freqs = count_to_freq( \%counts [, $pseudocount ] )
#
#        \@counts = [ [ n1, n2, n3, n4 ], [ n5, n6 ], ... [ nn1, nn2 ] ]
#        \%counts = { codon => count, ... }
#        \@freqs  = [ [ f1, f2, f3, f4 ], [ nf, f6 ], ... [ fn1, fn2 ] ]
#-----------------------------------------------------------------------------
sub count_to_freq
{
    my ( $cnts, $pseudocnt ) = @_;
    $cnts = codon_count_package( $cnts ) if ref $cnts eq 'HASH';
    return undef if ref $cnts  ne 'ARRAY';
    $pseudocnt ||= 0;

    [ map { my $i = 0;                          # elements in group
            my $n = $pseudocnt;
            foreach ( @$_ ) { $n += $_ ; $i++ } # total count in group
            $n ||= 1;                           #   or 1
            my $pci = $i ? $pseudocnt / $i : 0; # and per element pseudocount
            [ map { ( $_ + $pci ) / $n } @$_ ]  # used to make frequencies
          }
      @$cnts[ 0 .. 17 ]                         # for each group of counts
    ]
}


#-----------------------------------------------------------------------------
#  Compare packaged codon usage(s) to expected frequencies by chi-square.
#
#     ( $chisqr, $df, $n ) = packaged_codon_usage_chi_sqr( \@freqs, \@cnt )
#
#-----------------------------------------------------------------------------
sub packaged_codon_usage_chi_sqr
{
    my ( $freq, $cnt ) = @_;
    ref( $freq ) eq 'ARRAY' && ref( $cnt ) eq 'ARRAY'
        || die "packaged_codon_usage_chi_sqr requires two ARRAY references\n";

    my @c2 = @$cnt[ 0 .. 17 ];
    my ( $chisqr, $df, $total ) = (0, 0, 0);

    foreach my $fr ( @$freq )
    {
        my ($c, $d, $n) = gjostat::chi_square( $fr, shift @c2 );
        if ( $d && $n ) { $chisqr += $c; $df += $d; $total += $n }
    }

    ( $chisqr, $df, $total )
}


#-----------------------------------------------------------------------------
#  Compare codon counts to expected frequencies by chi-square.
#
#     ( $chisqr, $df, $n ) = count_vs_freq_chi_sqr( \@cnts, \@freqs )
#     ( $chisqr, $df, $n ) = count_vs_freq_chi_sqr( \%cnts, \@freqs )
#
#-----------------------------------------------------------------------------
sub count_vs_freq_chi_sqr
{
    my ( $cnt, $freq ) = @_;
    $cnt  = codon_count_package( $cnt  ) if ref $cnt  eq 'HASH';
    $freq = codon_count_package( $freq ) if ref $freq eq 'HASH';  # Silly, but would work

    return () if ref $cnt ne 'ARRAY' || ref $freq ne 'ARRAY';

    my ( $chisqr, $df, $total ) = (0, 0, 0);
    for ( my $i = 0; $i <= 17; $i++ )
    {
        my ( $c, $d, $n ) = gjostat::chi_square( $freq->[$i], $cnt->[$i] );
        if ( $d && $n ) { $chisqr += $c; $df += $d; $total += $n }
    }

    ( $chisqr, $df, $total )
}


#-----------------------------------------------------------------------------
#  Score a gene's codon usage against one or more sets of frequencies:
#
#     @scores = codon_usage_p_values( \@codon_counts, \@codon_freq_sets, \%options )
#
#  Options:
#
#     max_l    => $max_length  #  Limit length used in p-value
#     max_len  => $max_length
#
#-----------------------------------------------------------------------------

sub codon_usage_p_values
{
    my ( $cnts, $freqs, $opts ) = @_;

    $freqs = [ $freqs ] if ! ref $freqs->[0]->[0];  #  Allow one set of frequences

    $opts ||= {};
    my $max_l = $opts->{ max_l } || $opts->{ max_len } || $opts->{ max_length };

    my $scale = 1;
    if ( $max_l )
    {
        my $n = 0; foreach ( map { @$_ } @$cnts[ 0 .. 17 ] ) { $n += $_ }
        return () if ! $n;
	    $scale = ( $max_l / $n ) if ( $n > $max_l );
	}

    map { my ( $chisqr, $df, undef ) = count_vs_freq_chi_sqr(  $cnts, $_ );
          $df ? gjostat::chisqr_prob( $scale * $chisqr, $df ) : 1
        } @$freqs
}


#-----------------------------------------------------------------------------
#  A score based on the sum of the p-values for all gene counts compared
#  of a set of relative codon usage frequencies:
#
#     $score = codon_freq_score( \@freq, \@per_gene_counts, \%options )
#     $score = codon_freq_score( \@freq, \%per_gene_counts, \%options )
#
#  Options:
#
#     expon    => exponent  #  Return sum of P**exponent
#     exponent => exponent  #  Same as expon
#
#     p_val    => P-value   #  Return the count of genes with P >= P-value
#     p_value  => P-value   #  Same as p_val
#-----------------------------------------------------------------------------

sub codon_freq_score
{
    my ( $freq, $cnts, $opts ) = @_;

    $cnts = { map { $_ => $_ } @$cnts } if ref $cnts eq 'ARRAY';

    $opts ||= {};
    my $p_val = $opts->{ p_val } || $opts->{ p_value };
    my $expon = $opts->{ expon } || $opts->{ exponent };

    my $score = 0;
    foreach my $id ( keys %$cnts )
    {
        my ( $chisqr, $df, undef ) = count_vs_freq_chi_sqr( $cnts->{ $id }, $freq );
        if ( $df )
        {
            my $p = gjostat::chisqr_prob( $chisqr, $df );
            $score += $p_val ? ( $p >= $p_val ? 1 : 0 )  # P >= p_value?
                    : $expon ? $p**$expon                # sum( P ** exponent )
                    :          $p;                       # sum( P )
        }
    }

    return $score;
}


#-----------------------------------------------------------------------------
#  Format counts with amino acids separated by 2 spaces, and codons within the
#  amino acid separated by 1 space.  If an id is present, it is separated by a
#  tab.
#
#  naa1c1 naa1c2 naa1c3 naa1c4  naa2c1 naa2c2 ...  naa3c1 naa3c2 ...
#
#     report_counts(       \@packaged_counts )
#     report_counts(       \@packaged_counts, $id )
#     report_counts( \*FH, \@packaged_counts )
#     report_counts( \*FH, \@packaged_counts, $id )
#
#-----------------------------------------------------------------------------

sub report_counts
{
    my $fh = ( ref( $_[0] ) eq 'GLOB' ) ? shift : \*STDOUT;
    my ( $cnts, $id ) = @_;
    print $fh join( '  ', map { join( ' ', map { $_ || 0 } @$_ ) } @$cnts ),
              ( $id ? "\t$id" : () ),
              "\n";
}


#-----------------------------------------------------------------------------
#  Split counts with amino acids separated by 2 spaces, and codons within the
#  amino acid separated by 1 space.  If an id is present, it is separated by a
#  tab.
#
#       \@counts        = split_counts( $codon_count_string )
#     ( \@counts, $id ) = split_counts( $codon_count_string )
#
#-----------------------------------------------------------------------------

sub split_counts
{   my ( $string ) = shift;
    chomp $string;
    my ( $data, $id ) = split /\t/, $string;

    my $cnts = [ map { [ map { $_ + 0 } split / / ] } split /  /, $data ];

    wantarray ? ( $cnts, $id ) : $cnts;
}


#-----------------------------------------------------------------------------
#  Format frequences with amino acids separated by vertical bars, and
#  codons within the amino acid separated by commas:
#
#  faa1c1,faa1c2,faa1c3,faa1c4|faa2c1,faa2c2,...|faa3c1,faa3c2,...
#
#     report_frequencies(               \@freqs [, $title] )
#     report_frequencies( \*FH,         \@freqs [, $title] )
#     report_frequencies(       $score, \@freqs [, $title] )
#     report_frequencies( \*FH, $score, \@freqs [, $title] )
#
#-----------------------------------------------------------------------------

sub report_frequencies
{
    my $fh    = ( ref( $_[0] ) eq 'GLOB' ) ? shift : \*STDOUT;

    my @parts = ( ( ! ref( $_[0] ) ? shift : () ),
                  frequencies_as_string( shift ),
                  ( $_[0]          ? shift : () )
                );
    print $fh join( "\t", @parts ), "\n";
}


#-----------------------------------------------------------------------------
#  Format frequences with amino acids separated by vertical bars, and
#  codons within the amino acid separated by commas:
#
#  faa1c1,faa1c2,faa1c3,faa1c4|faa2c1,faa2c2,...|faa3c1,faa3c2,...
#
#   $codon_freq_string = frequencies_as_string( \@freqs )
#
#-----------------------------------------------------------------------------

sub frequencies_as_string
{
    join( "|", map { join( ",", map { sprintf "%7.5f", $_ } @$_ ) } @{$_[0]} );
}


#-----------------------------------------------------------------------------
#  Split frequences with amino acids separated by vertical bars, and
#  codons within the amino acid separated by commas:
#
#  faa1c1,faa1c2,faa1c3,faa1c4|faa2c1,faa2c2,...|faa3c1,faa3c2,...
#
#      \@freqs                         = split_frequencies( $codon_freq_string )
#    ( \@freqs, $score, $description ) = split_frequencies( $codon_freq_string )
#
#-----------------------------------------------------------------------------

sub split_frequencies
{
    local $_ = shift;
    s/\s+$//;
    s/^\s+//;
    my ( undef, $scr, $freq, undef, $descr ) = $_ =~ m/^(([\d.]*)\t)?(\d[\d.]*,\d[^\t]*)(\t([^\t]*))?$/;
    return wantarray ? ( undef, undef, undef ) : undef if ! $freq;
    my $freq2 = [ map { [ map { $_ + 0 } split /,/ ] } split /\|/, $freq ];
    wantarray ? ( $freq2, $scr, $descr ) : $freq2
}



#===============================================================================
#  Convert packaged codon counts into an array of amino acid counts, for one
#  or more genes:
#
#      @per_gene_aa_cnts = codon_counts_2_aa_counts( @per_gene_codon_cnts )
#
#===============================================================================
sub codon_counts_2_aa_counts
{
    map { [ map { my $ttl = 0; foreach ( @$_ ) { $ttl += $_ } $ttl } @$_ ] } @_
}


#===============================================================================
#  Produce a simulated set of codon counts matched to the length and amino
#  acid composition of the genes:
#
#    @per_gene_codon_cnt = simulate_genome( \@packaged_freqs, @per_gene_aa_cnt )
#
#===============================================================================

sub simulate_genome
{
    my $freqs = shift;
    ref $freqs eq 'ARRAY'
        and ref $freqs->[0] eq 'ARRAY'
        or return undef;

    #  Encapsulate the codon frequencies at covering segments on the
    #  interval between 0 and 1.  This will allow efficient access to
    #  codons with the desired frequencies.

    my @aa_info;
    my @extra = ( [1] ) x ( 20 - @$freqs );
    foreach my $aa_freq ( @$freqs, @extra )
    {
        #  Ensure that the frequencies are normalized:

        my $ttl = 0;
        foreach ( @$aa_freq ) { $ttl += $_ }
        my @aa_freq = $ttl ? map { $_ / $ttl      } @$aa_freq
                           : map { 1  / @$aa_freq } @$aa_freq;

        #  Create a tree of the intervals:

        my $i = 0;
        my @pairs = map { [ $i++, $_ ] } @aa_freq;

        push @aa_info, [ segment_new_tree( @pairs ), scalar @aa_freq ];
    }

    #  Process each gene:

    map { my $gene = $_;
          my @gene_data;
          my $aa_num = 0;
          foreach my $aa_cnt ( @$gene )
          {
              #  Initialize the codon counts so that we alwoys get a list
              #  of the correctl length:

              my ( $tree, $n_codon ) = @{ $aa_info[ $aa_num++ ] };
              my @aa_data = ( 0 ) x $n_codon;
              for ( my $i = 0; $i < $aa_cnt; $i++ )
              {
                  $aa_data[ segment_by_coord( rand(), $tree ) ]++;
              }

              push @gene_data, \@aa_data;
          }
          \@gene_data
        } @_
}


#===============================================================================
#  Modal codon usage calculation:
#
#            \@modal_freqs   = modal_codon_usage( \@gene_cnt_pkgs, \%options )
#  ( $score, \@modal_freqs ) = modal_codon_usage( \@gene_cnt_pkgs, \%options )
#
#  Options:
#
#      average   => boolean   #  include average as seed vertex
#      exponent  => float     #  P_value_exponent in optimization (D = 0.3)
#      max_steps => max_step  #  max simplex steps in optimization (D = 1e6)
#      n_top     => n         #  maximum vertices in simplex optimization (D = 100)
#      pseudo    => float     #  pseudo counts included in codon frequencies (D = 0.1)
#      root      => temp_file #  root name (D = evaluate_mode_temp_000000000)
#      verbose   => int       #  reporting interval for opt steps (D = never)
#      vertices  => int       #  minimum vertices in simplex optimization (D = 50)
#
#===============================================================================
sub modal_codon_usage
{
    my ( $counts, $options ) = @_;
    $options ||= {};

    my ( $rootkey ) = grep { m/root/i } keys %$options;
    my $root_name = $options->{ $rootkey }
                 || "evaluate_mode_temp_@{[sprintf '%09d',int(1e9*rand())]}";

    my ( $avgkey ) = grep { m/^av/i } keys %$options;
    my $average = $options->{ $avgkey } || 0;

    my ( $expkey ) = grep { m/exp/i } keys %$options;
    my $expon = $options->{ $expkey } || 0.3;

    my ( $extrakey ) = grep { m/extra/i } keys %$options;
    my $extra = $options->{ $extrakey } || 0;

    my ( $pseudokey ) = grep { m/pseudo/i } keys %$options;
    my $pseudo = $options->{ $pseudokey } || 0.1;

    my ( $stepkey ) = grep { m/step/i } keys %$options;
    my $maxstep = $options->{ $stepkey } || 1e6;

    my ( $verbkey ) = grep { m/verb/i } keys %$options;
    my $verbose = $options->{ $verbkey };

    my ( $vertkey ) = grep { m/vert/i } keys %$options;
    my $vertices = $options->{ $vertkey } || 50;

    my ( $topkey ) = grep { m/top/i } keys %$options;
    my $n_top = $options->{ $topkey } || ( $vertices > 100 ? $vertices : 100 );

    my $counts_file = "$root_name.counts";

    #  The optimization code uses the counts file

    # print STDERR "Writing file with @{[scalar @$counts]} gene codon counts.\n";
    open( CNT, ">$counts_file" )
        or print STDERR "modal_codon_usage() could not open '$counts_file' for writing.\n"
        and exit;
    foreach ( @$counts ) { report_counts( \*CNT, $_ ) }
    close CNT;

    my @freqs = map { count_to_freq( $_, $pseudo ) } @$counts;
    push @freqs, count_to_freq( sum_counts( $counts ), $pseudo ) if $average;

    my $scr_opts = { counts   => $counts_file,
                     exponent => $expon,
                     vertices => $vertices
                   };

    my @best_freqs = sort { $b->[0] <=> $a->[0] }
                     score_codon_frequencies( \@freqs, $scr_opts );

    splice @best_freqs, $n_top;

    #  n_top is number of actual to keep
    #  extra is number above min_vertices to explore

    my $opt_opts = { counts   => $counts_file,
                     exponent => $expon,
                     extra    => $extra,
                     # verbose  => 1,
                     vertices => $vertices
                   };
    
    my ( $score, $modal_freqs ) = optimize_frequencies( \@best_freqs, $opt_opts );
    
    unlink $counts_file;

    wantarray() ? ( $score, $modal_freqs ) : $modal_freqs
}


#===============================================================================
#  Score requencies relative to a set of gene counts:
#
#    @scored_freqs = score_codon_frequencies( \@freqs, \@counts, \%options )
#   \@scored_freqs = score_codon_frequencies( \@freqs, \@counts, \%options )
#    @scored_freqs = score_codon_frequencies( \@freqs,           \%options )
#   \@scored_freqs = score_codon_frequencies( \@freqs,           \%options )
#
#  Output is [ score, freqs ] pairs.
#
#  Codon counts for scoring can be supplied in the command, or can be supplied
#  by a file named in the options.  If both counts and a name are supplied,
#  this is used as the temporary file name.
#
#  Options:
#
#      count_file => file      #  name for the codon counts file
#      exponent   => expon     #  P-value exponent for scoring total P-value**expon (D = 0.3)
#      max_length => max_len   #  max gene codons in calculating P-values (D = unlim)
#      p_value    => p-value   #  cutoff for counting hits (D = 0.1)
#      verbose    => interval  #  reporting interval for count of genes scored (D = never)
#
#  By default, the score is the sum of exponentiated P-values
#===============================================================================
sub score_codon_frequencies
{
    my $freqs = shift;
    ref( $freqs ) eq 'ARRAY'
        or print STDERR 'score_codon_frequencies() called with invalid freqs\n'
        and exit;

    my $counts = ( ref($_[0]) eq 'ARRAY' ) ? shift : [];

    my $options = ( ref($_[0]) eq 'HASH' ) ? shift : {};

    my ( $cntkey ) = grep { m/count/i || m/file/i } keys %$options;
    my $cntfile = $options->{ $cntkey }
               || sprintf "score_codon_frequencies_tmp_%09d.counts", int( 1e9 * rand() );

    my ( $expkey ) = grep { m/exp/i } keys %$options;
    my $expon = $options->{ $expkey } || 0.3;

    my ( $lenkey ) = grep { m/len/i } keys %$options;
    my $max_l = $options->{ $lenkey };

    my ( $pvalkey ) = grep { m/p_v/i } keys %$options;
    my $p_val = $options->{ $pvalkey };

    my ( $verbkey ) = grep { m/verb/i } keys %$options;
    my $verbose = $options->{ $verbkey };

    my $save_cnt = -f $cntfile;
    if ( @$counts )
    {
        open( CNT, ">$cntfile" )
            or print STDERR "score_codon_frequencies() could not open '$cntfile' for writing.\n"
            and exit;
        foreach ( @$counts ) { report_counts( \*CNT, $_ ) }
        close CNT;
    }
    elsif ( ! $save_cnt )
    {
        print STDERR "score_codon_frequencies() called with neither count data or a count file.\n";
        exit;
    }

    #  Open the evaluation pipe:

    my $eval_cmd = join( ' ', 'codon_freq_eval_2',
                              ( $max_l ? sprintf( '-l %d',   $max_l ) : () ),
                              ( $p_val ? sprintf( '-p %.3e', $p_val ) : sprintf( '-e %.3e', $expon ) ),
                              $cntfile
                       );

    my( $pid, $rd, $wr );
    $pid = open2( $rd, $wr, $eval_cmd )
        or print STDERR "score_codon_frequencies() could not open evaluation pipe to:\n    '$eval_cmd'\n"
           and exit;
    { my $old = select $wr; $| = 1; select $old; }  #  Autoflush the write pipe

    my $ndone = 0;
    my @scored;
    foreach my $freq ( @$freqs )
    {
        #  If called with scored frequencies ( [$score, $freq] ), fix them.
        $freq = $freq->[1] if ( @$freq == 2 );
        push @scored, [ calc_freq_score( $wr, $rd, $freq ), $freq ];
        print STDERR "score_codon_frequencies: $ndone done.\n" if $verbose && ( (++$ndone % $verbose) == 0 );
    }

    close( $wr );
    close( $rd );
    waitpid $pid, 0;

    unlink $cntfile if ! $save_cnt;

    wantarray() ? @scored : \@scored
}


#===============================================================================
#  Optimize codon frequencies:
#
#              $freqs   = optimize_frequencies( \@freqs, \@counts, \%options )
#              $freqs   = optimize_frequencies( \@freqs,           \%options )
#    ( $score, $freqs ) = optimize_frequencies( \@freqs,           \%options )
#    ( $score, $freqs ) = optimize_frequencies( \@freqs, \@counts, \%options )
#
#  Options:
#
#      count_file   => file      #  name for the codon counts file
#      exponent     => expon     #  P-value exponent for scoring total
#                                #     P-value**expon (D = 0.3)
#      extra_vertices => extra   #  number of extra starting points to evaluate
#                                #     saving on the best
#      max_length   => max_len   #  max gene codons in calculating P-values
#                                #     (D = unlimited, changes not recommended)
#      max_steps    => max_step  #  max simplex steps in optimization (D = 1e6)
#      min_vertices => vertices  #  minimum number of vertices (D = 42)
#      p_value      => p-value   #  report number with p >= p-value for score
#      verbose      => interval  #  reporting interval for opt steps (D = never)
#
#===============================================================================
sub optimize_frequencies
{
    my $freqs = shift;
    ref( $freqs ) eq 'ARRAY'
        or print STDERR 'optimize_frequencies() called with invalid freqs\n'
        and exit;

    my $counts = ( ref($_[0]) eq 'ARRAY' ) ? shift : [];

    my $options = ( ref($_[0]) eq 'HASH' ) ? shift : {};

    my ( $cntkey ) = grep { m/count/i || m/file/i } keys %$options;
    my $cntfile = $options->{ $cntkey }
               || "optimize_frequencies_temp_@{[sprintf '%09d',int(1e9*rand())]}.counts";

    my ( $expkey ) = grep { m/exp/i } keys %$options;
    my $expon = $options->{ $expkey } || 0.3;

    my ( $extrakey ) = grep { m/extra/i } keys %$options;
    my $extra = $options->{ $extrakey } || 0;

    my ( $lenkey ) = grep { m/len/i } keys %$options;
    my $max_l = $options->{ $lenkey };

    my ( $stepkey ) = grep { m/step/i } keys %$options;
    my $maxstep = $options->{ $stepkey } || 1e6;

    my ( $verbkey ) = grep { m/verb/i } keys %$options;
    my $verbose = $options->{ $verbkey };

    my ( $vertkey ) = grep { m/vert/i } keys %$options;
    my $vertices = $options->{ $vertkey } || 42;

    my $save_cnt = -f $cntfile;
    if ( @$counts )
    {
        open( CNT, ">$cntfile" )
            or print STDERR "optimize_frequencies() could not open '$cntfile' for writing.\n"
            and exit;
        foreach ( @$counts ) { report_counts( \*CNT, $_ ) }
        close CNT;
    }
    elsif ( ! $save_cnt )
    {
        print STDERR "optimize_frequencies() called with neither count data or a count file.\n";
        exit;
    }

    #  Do we have >= 6 points?

    my @freqs = @$freqs;
    if ( @freqs < 6 )
    {
        print STDERR <<"End_of_Few_Points";

Cannot optimize codon usage frequencies with only @{[scalar @freqs]} points.
Consider using average codon usage.

End_of_Few_Points

        exit 1;
    }

    #  Open the evaluation pipe:

    my $eval_cmd = join( ' ', 'codon_freq_eval_2',
                              ( $max_l ? sprintf( '-l %d',   $max_l ) : () ),
                              sprintf( '-e %.3e', $expon ),
                              $cntfile
                       );

    my( $pid, $rd, $wr );
    $pid = open2( $rd, $wr, $eval_cmd )
        or print STDERR "optimize_frequencies() could not open evaluation pipe:\n    '$eval_cmd'\n"
           and exit;
    { my $old = select $wr; $| = 1; select $old; }  #  Autoflush the write pipe

    #  Score frequencies, if they are not already score-frequency pairs:

    if ( @{ $freqs[0] } != 2 )
    {
        @freqs = map { [ calc_freq_score( $wr, $rd, $_ ), $_ ] } @freqs;
    }

    my $nstep = 0;
    my @attempts;
    if ( $verbose )
    {
        foreach ( @freqs )
        {
            report_attempts( $nstep, $_->[0] ) if ++$nstep % $verbose == 0;
        }
    }

    #  Order them best to worst:

    @freqs = sort { $b->[0] <=> $a->[0] } @freqs;

    #  Do we need more vertices?  Make combinations of amino acid-specific
    #  compositions, drawn randomly from among the available points.  (This
    #  is way too easy.)

    my $np  = @freqs;               # number of preexisting points
    my $naa = @{ $freqs[0]->[1] };  # number of amino acids
    $vertices = @freqs if $vertices < @freqs;
    while ( @freqs < $vertices + $extra )
    {
        my $freq = random_freq_3( \@freqs, $naa );
        my $scr = calc_freq_score( $wr, $rd, $freq );
        report_attempts( $nstep, $scr ) if $verbose && ++$nstep % $verbose == 0;
        @freqs = sort { $b->[0] <=> $a->[0] } @freqs, [ $scr, $freq ];
    }

    splice @freqs, $vertices;

    # return $freqs[0]->[1];     #############################################

    #  Do a simplex optimization of the score by moving the orginal points
    #
    #  Try up to 4 options for test point.  Nothing is done to maintain
    #  the normalization of the frequencies, but round off error in the
    #  point movements is so small that this does not see to accumulate
    #  significantly.  The chi-square itself enforces normalization of the
    #  frequencies, so that cannot introduce systematic error.
    #
    #  Locations of points tested:
    #
    #     p0     p1    mean     p2      p3      p4
    #     0.0    0.5    1.0     1.5     2.0     2.4
    #           shrink         move    move    move
    #                           and             and
    #                          shrink          grow

    # my @step = ( 0.0, 0.4, 1.6, 2.0, 2.4 );
    my @step = ( 0.8, 0.5, 1.5, 2.0, 2.5 );

    my $done = 0;
    while ( ! $done )
    {
        #  Order points from worst to best:

        @freqs = sort { $a->[0] <=> $b->[0] } @freqs;
        my ( $wrst_scr, $wrst_pnt ) = @{ $freqs[  0 ] };  # Worst
        my ( $best_scr, $best_pnt ) = @{ $freqs[ -1 ] };  # Best

        #  Stop if there is no significant spread of scores;

        if ( ( $best_scr - $wrst_scr ) < 1e-6 * $best_scr ) { $done = 1; next }
        @attempts = ( ++$nstep );

        #  Try to improve one vertex, from worst to best:

        my $imprv_pnt = undef;
        while ( ( ! $imprv_pnt ) && ( ! $done ) )
        {
            my ( $scr, $p0 ) = @{ shift @freqs };
            push @attempts, $scr;
            my $mean_pnt = mean_point( map { $_->[1] } @freqs ); # p0 not in @freqs
            my $dir = subtract_points( $mean_pnt, $p0 );

            my $p1 = move_point( $p0, $dir, $step[1] );
            if ( $p1 )
            {
                my $s1 = calc_freq_score( $wr, $rd, $p1 );
                if ( $s1 > $scr ) { $scr = $s1; $imprv_pnt = $p1 }
                push @attempts, $s1;
            }

            my $p2 = move_point( $p0, $dir, $step[2] );
            if ( $p2 )
            {
                my $s2 = calc_freq_score( $wr, $rd, $p2 );
                push @attempts, $s2;
                if ( $s2 > $scr )
                {
                    $scr = $s2;
                    $imprv_pnt = $p2;

                    #  Only consider p3 if p2 is current best

                    my $p3 = move_point( $p0, $dir, $step[3] );
                    if ( $p3 )
                    {
                        my $s3 = calc_freq_score( $wr, $rd, $p3 );
                        push @attempts, $s3;
                        if ( $s3 > $scr )
                        {
                            $scr = $s3;
                            $imprv_pnt = $p3;

                            #  Only consider p4 if p3 is current best

                            my $p4 = move_point( $p0, $dir, $step[4] );
                            if ( $p4 )
                            {
                                my $s4 = calc_freq_score( $wr, $rd, $p4 );
                                push @attempts, $s4;
                                if ( $s4 > $scr ) { $scr = $s4; $imprv_pnt = $p4 }
                            }
                        }
                    }
                }
            }
            push @freqs, [ $scr, $imprv_pnt || $p0 ];    #  Final point
            report_attempts( @attempts ) if $verbose && $nstep % $verbose == 0;

            if  ( ! $imprv_pnt && $p0 eq $best_pnt )     #  Tried all points
            {
                #  Last ditch effort to recover the optimizaton.
                #  Take step most of the way to the average.

                $p0 = undef;
                while ( ! $imprv_pnt && ( $p0 ne $best_pnt ) )
                {
                    ( $scr, $p0 ) = @{ shift @freqs };
                    @attempts = ( '', $scr );
                    $mean_pnt = mean_point( map { $_->[1] } @freqs ); # p0 not in @freqs
                    $dir = subtract_points( $mean_pnt, $p0 );
                    $p1 = move_point( $p0, $dir, $step[0] );
                    if ( $p1 )
                    {
                        my $s1 = calc_freq_score( $wr, $rd, $p1 );
                        if ( $s1 > $scr ) { $scr = $s1; $imprv_pnt = $p1 }
                        push @attempts, $s1;
                    }
                    push @freqs, [ $scr, $imprv_pnt || $p0 ];
                    report_attempts( @attempts ) if $verbose && $nstep % $verbose == 0;
                }
                $done = 1 if ! $imprv_pnt;
            }

            if ( ! $imprv_pnt ) { @attempts = ( '' ) }  # Not a new step
        }

        $done = 1 if ( $nstep >= $maxstep );
    }

    close( $wr );
    close( $rd );
    waitpid $pid, 0;

    unlink $cntfile if ! $save_cnt;

	my ( $best ) = sort { $b->[0] <=> $a->[0] } @freqs;

    wantarray() ? @$best : $best->[1]
}


sub merge_new_freq
{
    my ( $freqs, $new ) = @_;
    my ( $i, $j ) = ( 0, @$freqs );
    my $k;
    while ( $i != $j )
    {
        my $k = int( ( $i+$j ) / 2 );
        if ( $new->[0] >= $freqs->[$k]->[0] ) { $j = $k } else { $i = $k-1 }
    }
    splice @$freqs, $i, 0, $new;
    $freqs
}


#-------------------------------------------------------------------------------
#  Four ways to generate a new set of frequencies based on the current ones:
#
#      $newfreq = random_freq_1( \@freqs, $n_aa )
#      $newfreq = random_freq_2( \@freqs, $n_aa )
#      $newfreq = random_freq_3( \@freqs, $n_aa )
#      $newfreq = random_freq_4( \@freqs, $n_aa )
#
#-------------------------------------------------------------------------------
#  Each amino acid drawn from one member
#-------------------------------------------------------------------------------
sub random_freq_1
{
    my ( $freq, $naa ) = @_;
    my $nf = @$freq;
    [ map { [ @{ $freq->[int($nf*rand())]->[1]->[$_] } ] } ( 0 .. $naa-1 ) ]
}

#-------------------------------------------------------------------------------
#  Each amino acid drawn from one member, biased toward low numbers
#-------------------------------------------------------------------------------
sub random_freq_2
{
    my ( $freq, $naa ) = @_;
    my $nf = @$freq;
    [ map { [ @{ $freq->[int($nf*rand()**2)]->[1]->[$_] } ] } ( 0 .. $naa-1 ) ]
}

#-------------------------------------------------------------------------------
#  Each amino acid as average of 2 others
#-------------------------------------------------------------------------------
sub random_freq_3
{
    mean_2_points( random_freq_1( @_ ), random_freq_1( @_ ) );
}

#-------------------------------------------------------------------------------
#  Each amino acid as average of 2 others, biased toward low numbers
#-------------------------------------------------------------------------------
sub random_freq_4
{
    mean_2_points( random_freq_2( @_ ), random_freq_2( @_ ) );
}


#-------------------------------------------------------------------------------
#  Local calculation of a frequence score using an open bidirectional pipe:
#
#      $score = calc_freq_score( $wr, $rd, $freq )
#
#-------------------------------------------------------------------------------
sub calc_freq_score
{
    my ( $wr, $rd, $freq ) = @_;
    ( ref $freq eq 'ARRAY' and ref $freq->[0] eq 'ARRAY' )
        or print STDERR Dumper( $wr, $rd, $freq )
           and croak;
    print $wr join( " ", map { @$_ } @$freq ), "\n";
    <$rd> + 0
}


#  Output a set of frequencies in a reasonably human friendly form:

sub debug_freq
{
    my ( $freq, $lbl ) = @_;
    if ( $lbl ) { print STDERR "$lbl\n" }
    print STDERR join( "\n", map { join( ", ", map { sprintf "%7.4f", $_ } @$_ ) } @$freq ), "\n\n";
}


#-------------------------------------------------------------------------------
#  Report a log of simplex point evaluation scores:
#
#    report_attempts( $step, @scores )
#
#-------------------------------------------------------------------------------
sub report_attempts
{
    my $step = shift;
    print STDERR join( ' ', sprintf( '%6s', $step ), map { sprintf( '%11.6f', $_ ) } @_ ), "\n";
}


#-------------------------------------------------------------------------------
#  Find the mean of a list of points:
#
#    \@point = mean_point( \@point1, \@point2, ... )
#
#-------------------------------------------------------------------------------
sub mean_point
{
    my $p = copy_point( shift );
    my $n = 1;
    foreach ( @_ ) { $p = add_to_point( $p, $_ ); $n++ }
    scale_point( $p, 1/$n );
}


#-------------------------------------------------------------------------------
#  Make a copy of a point:
#-------------------------------------------------------------------------------
sub copy_point { [ map { [ @$_ ] } @{$_[0]} ] }


#-------------------------------------------------------------------------------
#  Scale a copy of a point:
#
#      \@point2 = scaled_point( \@point, $factor )
#
#-------------------------------------------------------------------------------
sub scaled_point
{
    my ( $point, $factor ) = @_;
    [ map { [ map { $factor * $_ } @$_ ] } @$point ]
}


#-------------------------------------------------------------------------------
#  Rescale a point in place:
#
#      \@point = scale_point( \@point, $factor )
#
#-------------------------------------------------------------------------------
sub scale_point
{
    my ( $point, $factor ) = @_;
    foreach ( @$point ) { foreach ( @$_ ) { $_ *= $factor } }
    $point
}


#-------------------------------------------------------------------------------
#  Add to a point in place:
#
#      \@point = add_to_point( \@point, \@delta )
#
#-------------------------------------------------------------------------------
sub add_to_point
{
    my ( $point, $delta ) = @_;
    my $d = copy_point( $delta );    # So that we do not destroy the original
    my $dp;
    foreach ( @$point ) { $dp = shift @$d; foreach ( @$_ ) { $_ += shift @$dp } }
    $point
}


#-------------------------------------------------------------------------------
#  Add two points:
#
#      \@point = add_points( \@point1, \@point2 )
#
#-------------------------------------------------------------------------------
sub add_points
{
    my ( $point1, $point2 ) = @_;
    my $p2 = copy_point( $point2 );    # So that we do not destroy the original
    my $p2p;
    [ map { $p2p = shift @$p2; [ map { $_ + (shift @$p2p) } @$_ ] } @$point1 ]
}


#-------------------------------------------------------------------------------
#  Mean of two points:
#
#      \@point = mean_2_points( \@point1, \@point2 )
#
#-------------------------------------------------------------------------------
sub mean_2_points
{
    my ( $point1, $point2 ) = @_;
    my $p2 = copy_point( $point2 );    # So that we do not destroy the original
    my $p2p;
    [ map { $p2p = shift @$p2; [ map { 0.5 * ( $_ + (shift @$p2p) ) } @$_ ] } @$point1 ]
}


#-------------------------------------------------------------------------------
#  Subtract two points:
#
#      \@point = subtract_points( \@point1, \@point2 )
#
#-------------------------------------------------------------------------------
sub subtract_points
{
    my ( $point1, $point2 ) = @_;
    my $p2 = copy_point( $point2 );    # So that we do not destroy the original
    my $p2p;
    [ map { $p2p = shift @$p2; [ map { $_ - (shift @$p2p) } @$_ ] } @$point1 ]
}


#-------------------------------------------------------------------------------
#  Move a point by a scaled direction vector
#
#    $point = move_point( $point0, $direction_vector, $distance )
#
#-------------------------------------------------------------------------------
sub move_point
{
    my ( $p0, $dir, $dist ) = @_;
    legal_point( add_points( $p0, scaled_point( $dir, $dist ) ) );
}


#-------------------------------------------------------------------------------
#  Check legality of composition of a point in codon frequency space:
#
#     $point = legal_point( \@point )
#
#-------------------------------------------------------------------------------
sub legal_point
{
    foreach ( @{ $_[0] } )
    {
        foreach ( @$_ ) { $_ >= 0 && $_ <= 1 || return undef }
    }
    $_[0]
}


#===============================================================================
#  Subroutines for codon frequency distances:
#
#    $distance = codon_freq_distance( \@freq1, \@freq2, $type )  # D = type 2
#    $distance = codon_freq_distance_1( \@freq1, \@freq2 )
#    $distance = codon_freq_distance_2( \@freq1, \@freq2 )
#    $distance = codon_freq_distance_3( \@freq1, \@freq2 )
#
#  The type 1 distance calculation treats each codon equally.  The difference in
#  relative codon usage for each codon is squared, and then summed over codons.
#                 ____
#                 \
#  distance1**2 =  |   ( f1(c) - f2(c) )**2
#                 /___
#               c = codons
#
#  where fn(c) is the frequency of codon c in frequency set n.
#
#  The type 2 distance calculation proceeds by amino acid.  For a given amino
#  acid, the sum of absolute differenes of codon frequencies is squared, and the
#  results are summed over amino acids.
#                     ____             ____
#                 1   \                \
#  distance2**2 = - *  |            (   |    abs( f1(c) - f2(c) ) )**2
#                 4   /___             /___
#                   a = amino acids  c = codons for amino acid a
#
#  The type 3 distance calculation treats each codon equally.  The distance
#  is the absolute difference in frequencies, summed over codons.  That is,
#  it is a Manhatten metric.
#              ____
#              \
#  distance3 =  |   abs( f1(c) - f2(c) )
#              /___
#            c = codons
#
#
#  Type 2 distances provide more equal treatment of amino acids with
#  different uses of codons.  N.b.,
#  ----------------------------------------------------------------
#           Codon usage for leucine         Distance**2 for leucine
#         ----------------------------      -----------------------
#         TTA  TTG  TCT  TCC  TCA  TCG        Type 1      Type 2
#  ----------------------------------------------------------------
#  set1    1    0    0    0    0    0            2          1
#  set2    0    1    0    0    0    0
#
#  set1    1    0    0    0    0    0           6/5         1
#  set2    0   1/5  1/5  1/5  1/5  1/5
#
#  set1   1/2  1/2   0    0    0    0           3/4         1
#  set2    0    0   1/4  1/4  1/4  1/4
#
#  set1   1/3  1/3  1/3   0    0    0           2/3         1
#  set2    0    0    0   1/3  1/3  1/3
#  ----------------------------------------------------------------
#
#  Since all of these scenarios have completely disjoint codon usages
#  between the two sets of frequencies, it makes sense that they have the
#  same score.
#
#===============================================================================

#  Allow method as 3rd arg:

sub codon_freq_distance
{
    my $type = $_[2] || 2;

    return  $type == 2 ? codon_freq_distance_2( @_[0,1] )
          : $type == 1 ? codon_freq_distance_1( @_[0,1] )
          : $type == 3 ? codon_freq_distance_3( @_[0,1] )
          : undef
}


sub codon_freq_distance_1
{
    my ( $f1, $f2 ) = @_;
    ref( $f1 ) eq 'ARRAY' && ref( $f2 ) eq 'ARRAY'
        or print STDERR "freq_distance_1 called with inappropriate args.\n"
        and return undef;

    my $dist_sq = 0;
    for ( my $aa = 0; $aa < @$f1; $aa++ )
    {
        $dist_sq += aa_freq_dist_sq_1( $f1->[$aa], $f2->[$aa] );
    }

    sqrt( $dist_sq )
}

sub aa_freq_dist_sq_1
{
    my ( $f1, $f2 ) = @_;

    my $ttl = 0;
    for ( my $codon = 0; $codon < @$f1; $codon++ )
    {
        $ttl += ( $f1->[$codon] - $f2->[$codon] ) ** 2;
    }

    $ttl
}

sub codon_freq_distance_2
{
    my ( $f1, $f2 ) = @_;
    ref( $f1 ) eq 'ARRAY' && ref( $f2 ) eq 'ARRAY'
        or print STDERR "freq_distance_2 called with inappropriate args.\n"
        and return undef;

    my $dist_sq = 0;
    for ( my $aa = 0; $aa < @$f1; $aa++ )
    {
        $dist_sq += aa_freq_dist_sq_2( $f1->[$aa], $f2->[$aa] );
    }

    sqrt( $dist_sq )
}

sub aa_freq_dist_sq_2
{
    my ( $f1, $f2 ) = @_;

    my $ttl = 0;
    for ( my $codon = 0; $codon < @$f1; $codon++ )
    {
        $ttl += abs( $f1->[$codon] - $f2->[$codon] );
    }

    0.25 * $ttl ** 2
}

sub codon_freq_distance_3
{
    my ( $f1, $f2 ) = @_;
    ref( $f1 ) eq 'ARRAY' && ref( $f2 ) eq 'ARRAY'
        or print STDERR "freq_distance_3 called with inappropriate args.\n"
        and return undef;

    my $dist_sq = 0;
    for ( my $aa = 0; $aa < @$f1; $aa++ )
    {
        $dist_sq += aa_freq_dist_3( $f1->[$aa], $f2->[$aa] );
    }

    $dist_sq
}

sub aa_freq_dist_3
{
    my ( $f1, $f2 ) = @_;

    my $ttl = 0;
    for ( my $codon = 0; $codon < @$f1; $codon++ )
    {
        $ttl += abs( $f1->[$codon] - $f2->[$codon] );
    }
}


#===============================================================================
#  For each frequency in the arguements, find the position along the
#  $freq0 to $freq1 vector that is closest to the frequencies point.
#  Coordinates along vector are measured with $freq0 being 0, and $freq1
#  being 1.
#
#     @projections = project_on_freq_vector_by_dist( \@freq_0, \@freq_1,   \@freq1, \@freq2, ...   )
#     @projections = project_on_freq_vector_by_dist( \@freq_0, \@freq_1, [ \@freq1, \@freq2, ... ] )
#
#  The returned projections are pairs composed of the position on the axis and
#  distance from the axis to the gene frequency.
#===============================================================================

sub project_on_freq_vector_by_dist
{
    my $freq_0 = shift;  #  Frequencies at point 0
    my $freq_1 = shift;  #  Frequencies at point 1
    return () if ! @_;
    my $d_freq_d_x = subtract_points( $freq_1, $freq_0 );
    return undef if zero_vector( $d_freq_d_x );

    #  Find the lower and upper bounds of x for each frequency:

    my @d_freq_d_x  = map { @$_ } @$d_freq_d_x;

    my @min_and_max = map { min_and_max_x( $_, ( shift @d_freq_d_x ) ) }
                      map { @$_ }
                      @$freq_0[ 0 .. 17 ];

    #  Find the maximum of the lower bounds, and minimum of the upper bounds:

    my ( $min_x, $max_x ) = ( -1e9, +1e9 );
    foreach ( @min_and_max )
    {
        $min_x = $_->[0] if $_->[0] > $min_x;
        $max_x = $_->[1] if $_->[1] < $max_x;
    }

    #  For each frequency in the arguements, find the position along the
    #  $freq0 to $freq1 vector that is closest to the point:

    my @projections = map { project_by_min_distance( $freq_0, $d_freq_d_x, $_, $min_x, $max_x ) }
                      ( ref $_[0]->[0]->[0] ? @{$_[0]} : @_ );

    wantarray ? @projections : \@projections
}


sub project_by_min_distance
{
    my ( $freq_0, $d_freq_d_x, $freq, $min_x, $max_x ) = @_;
    my $n_step = 16;
    my $inc = ( $max_x - $min_x ) / $n_step;
    my ( $x, $f );
    my @dists = map { $x = $inc * $_ + $min_x;
                      $f = move_point( $freq_0, $d_freq_d_x, $x );
                      [ $x, codon_freq_distance_2( $freq, $f ) ]
                    }
                ( 1 .. ($n_step-1) );  #  Leave empty space at ends

    my $x_and_dist = ( sort { $a->[1] <=> $b->[1] } @dists )[0];

    #  Divide and conquer search centered on current best point

    while ( $inc > 0.001 )
    {
        $inc *= 0.5;
        @dists = ( $x_and_dist,
                   map { $f = move_point( $freq_0, $d_freq_d_x, $_ );
                         [ $_, codon_freq_distance_2( $freq, $f ) ]
                       }
                   ( $x_and_dist->[0] - $inc, $x_and_dist->[0] + $inc )
                 );
        $x_and_dist = ( sort { $a->[1] <=> $b->[1] } @dists )[0];
    }

    $x_and_dist
}


#===============================================================================
#  For each frequency in the arguements, find the position along the
#  $freq0 to $freq1 vector that gives the smallest chi square value.
#  Coordinates along vector are measured with $freq0 being 0, and $freq1
#  being 1.
#
#     @projections = project_on_freq_vector_by_chi_sqr( \@freq_0, \@freq_1,   \@cnts1, \@cnts2, ...   )
#     @projections = project_on_freq_vector_by_chi_sqr( \@freq_0, \@freq_1, [ \@cnts1, \@cnts2, ... ] )
#
#  The returned projections are quartets composed of the position on the axis
#  the chi square value, the degrees of freedom and the number of different
#  amino acids.
#===============================================================================

sub project_on_freq_vector_by_chi_sqr
{
    my $freq_0 = shift;  #  Frequencies at point 0
    my $freq_1 = shift;  #  Frequencies at point 1
    return () if ! @_;
    my $d_freq_d_x = subtract_points( $freq_1, $freq_0 );
    return undef if zero_vector( $d_freq_d_x );

    #  Find the lower and upper bounds of x for each frequency:

    my @d_freq_d_x  = map { @$_ } @$d_freq_d_x;

    my @min_and_max = map { min_and_max_x( $_, ( shift @d_freq_d_x ) ) }
                      map { @$_ }
                      @$freq_0[ 0 .. 17 ];

    #  Find the maximum of the lower bounds, and minimum of the upper bounds:

    my ( $min_x, $max_x ) = ( -1e9, +1e9 );
    foreach ( @min_and_max )
    {
        $min_x = $_->[0] if $_->[0] > $min_x;
        $max_x = $_->[1] if $_->[1] < $max_x;
    }

    #  For each set of counts in the arguements, find the position along the
    #  $freq0 to $freq1 vector that has smallest chi square:

    my @projections = map { project_by_min_chi_sqr( $freq_0, $d_freq_d_x, $_, $min_x, $max_x ) }
                      ( ref $_[0]->[0]->[0] ? @{$_[0]} : @_ );

    wantarray ? @projections : \@projections
}


sub project_by_min_chi_sqr
{
    my ( $freq_0, $d_freq_d_x, $counts, $min_x, $max_x ) = @_;
    my $n_step = 16;
    my $inc = ( $max_x - $min_x ) / $n_step;
    my ( $x, $f );
    my @chi_sqrs = map { $x = $inc * $_ + $min_x;
                         $f = move_point( $freq_0, $d_freq_d_x, $x );
                         [ $x, count_vs_freq_chi_sqr( $counts, $f ) ]
                       }
                   ( 1 .. ($n_step-1) );  #  Leave empty space at ends

    my $x_and_chi_sqr = ( sort { $a->[1] <=> $b->[1] } @chi_sqrs )[0];

    #  Divide and conquer search centered on current best point

    while ( $inc > 0.001 )
    {
        $inc *= 0.5;
        @chi_sqrs = ( $x_and_chi_sqr,
                      map { $f = move_point( $freq_0, $d_freq_d_x, $_ );
                            [ $_, count_vs_freq_chi_sqr( $counts, $f ) ]
                          }
                      ( $x_and_chi_sqr->[0] - $inc, $x_and_chi_sqr->[0] + $inc )
                    );
        $x_and_chi_sqr = ( sort { $a->[1] <=> $b->[1] } @chi_sqrs )[0];
    }

    $x_and_chi_sqr
}


sub projection_confidence_interval
{
    my ( $freq_0, $d_freq_d_x, $cnts, $x_opt, $min_x, $max_x, $options ) = @_;

    my $options ||= {};
    my ( $p_opt ) = codon_usage_p_values( $cnts, [ move_point( $freq_0, $d_freq_d_x, $x_opt ) ], $options );

    return ( $min_x, $max_x, $p_opt ) if $p_opt < 1e-10;
    my $p_target = 0.05 * $p_opt;

    #  Divide and conquer search for points with desired p-value

    my $inc = 0.5 * ( $max_x - $x_opt );
    my $x_max = $x_opt + $inc;
    while ( $inc > 0.001 )
    {
        $inc *= 0.5;
        my ( $p ) = codon_usage_p_values( $cnts, [ move_point( $freq_0, $d_freq_d_x, $x_max ) ], $options );
        $x_max += ( $p_target <=> $p ) * $inc;
    }

    my $inc = 0.5 * ( $x_opt - $min_x );
    my $x_min = $x_opt - $inc;
    while ( $inc > 0.001 )
    {
        $inc *= 0.5;
        my ( $p ) = codon_usage_p_values( $cnts, [ move_point( $freq_0, $d_freq_d_x, $x_min ) ], $options );
        $x_min -= ( $p_target <=> $p ) * $inc;
    }

    ( $x_min, $x_max, $p_opt )
}


sub zero_vector
{
    foreach ( @{ shift @_ }[ 0 .. 17 ] ) { foreach ( @$_ ) { return 0 if $_ } }
    return 1;
}


#  Find the minimum and maximum x coordinates at which a frequency is in
#  the range $limit <= $freq <= ( 1 - $limit ), where $f_0 is the frequency
#  at x = 0 and $df_dx is the change in frequency per unit x.

sub min_and_max_x
{
    my ( $f_0, $df_dx, $limit ) = @_;
    return () if ! $df_dx;
    $limit ||= 1e-3;
    my $x_0 = (      $limit - $f_0 ) / $df_dx;
    my $x_1 = (  1 - $limit - $f_0 ) / $df_dx;

    ( $df_dx > 0 ) ? [ $x_0, $x_1 ] : [ $x_1, $x_0 ]
}


1;
